@@ -6,25 +6,28 @@ use std::io::Write;
 use bytelines::ByteLines;
 
 use crate::ansi;
+use crate::config::delta_unreachable;
 use crate::config::Config;
 use crate::features;
-use crate::handlers;
+use crate::handlers::hunk_header::ParsedHunkHeader;
+use crate::handlers::{self, merge_conflict};
 use crate::paint::Painter;
 use crate::style::DecorationStyle;
 
 #[derive(Clone, Debug, PartialEq)]
 pub enum State {
-    CommitMeta,                    // In commit metadata section
-    FileMeta, // In diff metadata section, between (possible) commit metadata and first hunk
-    HunkHeader(String, String), // In hunk metadata line (line, raw_line)
-    HunkZero, // In hunk; unchanged line
-    HunkMinus(Option<String>), // In hunk; removed line (raw_line)
-    HunkPlus(Option<String>), // In hunk; added line (raw_line)
+    CommitMeta,                                             // In commit metadata section
+    DiffHeader(DiffType), // In diff metadata section, between (possible) commit metadata and first hunk
+    HunkHeader(DiffType, ParsedHunkHeader, String, String), // In hunk metadata line (diff_type, parsed, line, raw_line)
+    HunkZero(DiffType, Option<String>), // In hunk; unchanged line (prefix, raw_line)
+    HunkMinus(DiffType, Option<String>), // In hunk; removed line (diff_type, raw_line)
+    HunkPlus(DiffType, Option<String>), // In hunk; added line (diff_type, raw_line)
+    MergeConflict(MergeParents, merge_conflict::MergeConflictCommit),
     SubmoduleLog, // In a submodule section, with gitconfig diff.submodule = log
     SubmoduleShort(String), // In a submodule section, with gitconfig diff.submodule = short
     Blame(String, Option<String>), // In a line of `git blame` output (commit, repeat_blame_line).
-    GitShowFile, // In a line of `git show $revision:./path/to/file.ext` output
-    Grep,     // In a line of `git grep` output
+    GitShowFile,  // In a line of `git show $revision:./path/to/file.ext` output
+    Grep,         // In a line of `git grep` output
     Unknown,
     // The following elements are created when a line is wrapped to display it:
     HunkZeroWrapped,  // Wrapped unchanged line
@@ -32,6 +35,39 @@ pub enum State {
     HunkPlusWrapped,  // Wrapped added line
 }
 
+#[derive(Clone, Debug, PartialEq)]
+pub enum DiffType {
+    Unified,
+    // https://git-scm.com/docs/git-diff#_combined_diff_format
+    Combined(MergeParents, InMergeConflict),
+}
+
+#[derive(Clone, Debug, PartialEq)]
+pub enum MergeParents {
+    Number(usize),  // Number of parent commits == (number of @s in hunk header) - 1
+    Prefix(String), // Hunk line prefix, length == number of parent commits
+    Unknown,
+}
+
+#[derive(Clone, Debug, PartialEq)]
+pub enum InMergeConflict {
+    Yes,
+    No,
+}
+
+impl DiffType {
+    pub fn n_parents(&self) -> usize {
+        use DiffType::*;
+        use MergeParents::*;
+        match self {
+            Combined(Prefix(prefix), _) => prefix.len(),
+            Combined(Number(n_parents), _) => *n_parents,
+            Unified => 1,
+            Combined(Unknown, _) => delta_unreachable("Number of merge parents must be known."),
+        }
+    }
+}
+
 #[derive(Debug, PartialEq)]
 pub enum Source {
     GitDiff,     // Coming from a `git diff` command
@@ -42,10 +78,10 @@ pub enum Source {
 // Possible transitions, with actions on entry:
 //
 //
-// | from \ to   | CommitMeta  | FileMeta    | HunkHeader  | HunkZero    | HunkMinus   | HunkPlus |
+// | from \ to   | CommitMeta  | DiffHeader  | HunkHeader  | HunkZero    | HunkMinus   | HunkPlus |
 // |-------------+-------------+-------------+-------------+-------------+-------------+----------|
 // | CommitMeta  | emit        | emit        |             |             |             |          |
-// | FileMeta    |             | emit        | emit        |             |             |          |
+// | DiffHeader  |             | emit        | emit        |             |             |          |
 // | HunkHeader  |             |             |             | emit        | push        | push     |
 // | HunkZero    | emit        | emit        | emit        | emit        | push        | push     |
 // | HunkMinus   | flush, emit | flush, emit | flush, emit | flush, emit | push        | push     |
@@ -58,8 +94,8 @@ pub struct StateMachine<'a> {
     pub source: Source,
     pub minus_file: String,
     pub plus_file: String,
-    pub minus_file_event: handlers::file_meta::FileEvent,
-    pub plus_file_event: handlers::file_meta::FileEvent,
+    pub minus_file_event: handlers::diff_header::FileEvent,
+    pub plus_file_event: handlers::diff_header::FileEvent,
     pub diff_line: String,
     pub painter: Painter<'a>,
     pub config: &'a Config,
@@ -70,7 +106,7 @@ pub struct StateMachine<'a> {
     // a file is renamed with changes, both are present, and we rely on the following variables to
     // avoid emitting the file meta header line twice (#245).
     pub current_file_pair: Option<(String, String)>,
-    pub handled_file_meta_header_line_file_pair: Option<(String, String)>,
+    pub handled_diff_header_header_line_file_pair: Option<(String, String)>,
     pub blame_commit_colors: HashMap<String, String>,
 }
 
@@ -90,11 +126,11 @@ impl<'a> StateMachine<'a> {
             source: Source::Unknown,
             minus_file: "".to_string(),
             plus_file: "".to_string(),
-            minus_file_event: handlers::file_meta::FileEvent::NoEvent,
-            plus_file_event: handlers::file_meta::FileEvent::NoEvent,
+            minus_file_event: handlers::diff_header::FileEvent::NoEvent,
+            plus_file_event: handlers::diff_header::FileEvent::NoEvent,
             diff_line: "".to_string(),
             current_file_pair: None,
-            handled_file_meta_header_line_file_pair: None,
+            handled_diff_header_header_line_file_pair: None,
             painter: Painter::new(writer, config),
             config,
             blame_commit_colors: HashMap::new(),
@@ -112,15 +148,20 @@ impl<'a> StateMachine<'a> {
                 self.source = detect_source(&self.line);
             }
 
+            // Every method named handle_* must return std::io::Result<bool>.
+            // The bool indicates whether the line has been handled by that
+            // method (in which case no subsequent handlers are permitted to
+            // handle it).
             let _ = self.handle_commit_meta_header_line()?
                 || self.handle_diff_stat_line()?
-                || self.handle_file_meta_diff_line()?
-                || self.handle_file_meta_minus_line()?
-                || self.handle_file_meta_plus_line()?
+                || self.handle_diff_header_diff_line()?
+                || self.handle_diff_header_minus_line()?
+                || self.handle_diff_header_plus_line()?
                 || self.handle_hunk_header_line()?
-                || self.handle_file_meta_misc_line()?
+                || self.handle_diff_header_misc_line()?
                 || self.handle_submodule_log_line()?
                 || self.handle_submodule_short_line()?
+                || self.handle_merge_conflict_line()?
                 || self.handle_hunk_line()?
                 || self.handle_git_show_file_line()?
                 || self.handle_blame_line()?
@@ -150,9 +191,9 @@ impl<'a> StateMachine<'a> {
         }
         if self.config.max_line_length > 0
             && self.raw_line.len() > self.config.max_line_length
-            // We must not truncate ripgrep --json output
-            // TODO: An alternative might be to truncate `line` but retain
-            // `raw_line` untruncated?
+            // Do not truncate long hunk headers
+            && !self.raw_line.starts_with("@@")
+            // Do not truncate ripgrep --json output
             && !self.raw_line.starts_with('{')
         {
             self.raw_line = ansi::truncate_str(
@@ -166,12 +207,14 @@ impl<'a> StateMachine<'a> {
     }
 
     /// Skip file metadata lines unless a raw diff style has been requested.
-    fn should_skip_line(&self) -> bool {
-        self.state == State::FileMeta && self.should_handle() && !self.config.color_only
+    pub fn should_skip_line(&self) -> bool {
+        matches!(self.state, State::DiffHeader(_))
+            && self.should_handle()
+            && !self.config.color_only
     }
 
     /// Emit unchanged any line that delta does not handle.
-    fn emit_line_unchanged(&mut self) -> std::io::Result<bool> {
+    pub fn emit_line_unchanged(&mut self) -> std::io::Result<bool> {
         self.painter.emit()?;
         writeln!(
             self.painter.writer,
@@ -207,7 +250,11 @@ pub fn format_raw_line<'a>(line: &'a str, config: &Config) -> Cow<'a, str> {
 /// * git diff
 /// * diff -u
 fn detect_source(line: &str) -> Source {
-    if line.starts_with("commit ") || line.starts_with("diff --git ") {
+    if line.starts_with("commit ")
+        || line.starts_with("diff --git ")
+        || line.starts_with("diff --cc ")
+        || line.starts_with("diff --combined ")
+    {
         Source::GitDiff
     } else if line.starts_with("diff -u")
         || line.starts_with("diff -ru")
