@@ -1,7 +1,7 @@
 use std::borrow::Cow;
 use std::collections::{HashMap, HashSet};
 use std::path::Path;
-use sysinfo::{Pid, Process, ProcessExt, SystemExt};
+use sysinfo::{Pid, Process, ProcessExt, ProcessRefreshKind, SystemExt};
 
 use lazy_static::lazy_static;
 
@@ -30,7 +30,7 @@ pub fn calling_process() -> Option<Cow<'static, CallingProcess>> {
     }
     #[cfg(test)]
     {
-        determine_calling_process().map(|proc| Cow::Owned(proc))
+        determine_calling_process().map(Cow::Owned)
     }
 }
 
@@ -275,15 +275,11 @@ trait ProcessInterface {
         self.refresh_process(sibling_pid).then(|| ())?;
         self.process(sibling_pid)
     }
-    fn find_sibling_process<F, T>(&mut self, pid: Pid, extract_args: F) -> Option<T>
+    fn find_sibling_in_refreshed_processes<F, T>(&mut self, pid: Pid, extract_args: &F) -> Option<T>
     where
         F: Fn(&[String]) -> ProcessArgs<T>,
         Self: Sized,
     {
-        self.refresh_processes();
-
-        let this_start_time = self.process(pid)?.start_time();
-
         /*
 
         $ start_blame_of.sh src/main.rs | delta
@@ -302,6 +298,8 @@ trait ProcessInterface {
 
         */
 
+        let this_start_time = self.process(pid)?.start_time();
+
         let mut pid_distances = HashMap::<Pid, usize>::new();
         let mut collect_parent_pids = |pid, distance| {
             pid_distances.insert(pid, distance);
@@ -349,7 +347,8 @@ impl ProcessInterface for ProcInfo {
         std::process::id() as Pid
     }
     fn refresh_process(&mut self, pid: Pid) -> bool {
-        self.info.refresh_process(pid)
+        self.info
+            .refresh_process_specifics(pid, ProcessRefreshKind::new())
     }
     fn process(&self, pid: Pid) -> Option<&Self::Out> {
         self.info.process(pid)
@@ -358,7 +357,8 @@ impl ProcessInterface for ProcInfo {
         self.info.processes()
     }
     fn refresh_processes(&mut self) {
-        self.info.refresh_processes()
+        self.info
+            .refresh_processes_specifics(ProcessRefreshKind::new())
     }
 }
 
@@ -406,7 +406,8 @@ where
     started, so no command line can be parsed. Same if the data was piped from an input file.
 
     There might also be intermediary scripts in between or piped input with a gap in pids or (rarely)
-    randomized pids, so check all processes for the closest match in the process tree.
+    randomized pids, so check processes for the closest match in the process tree.
+    The size of this process tree can be reduced by only refreshing selected processes.
 
     100 /usr/bin/some-terminal-emulator
     124  \_ -shell
@@ -425,7 +426,27 @@ where
     567  |       \_ less --RAW-CONTROL-CHARS --quit-if-one-screen
 
     */
-    info.find_sibling_process(my_pid, extract_args)
+
+    let pid_range = my_pid.saturating_sub(10)..my_pid.saturating_add(20);
+    for p in pid_range.clone() {
+        // Processes which were not refreshed do not exist for sysinfo, so by selectively
+        // letting it know about processes the `find_sibling..` function will only
+        // consider these.
+        info.refresh_process(p);
+    }
+
+    match info.find_sibling_in_refreshed_processes(my_pid, &extract_args) {
+        None => {
+            #[cfg(test)]
+            {
+                info.refresh_processes();
+                info.find_sibling_in_refreshed_processes(my_pid, &extract_args)
+            }
+            #[cfg(not(test))]
+            None
+        }
+        some => some,
+    }
 }
 
 // Walk up the process tree, calling `f` with the pid and the distance to `starting_pid`.
@@ -487,10 +508,10 @@ pub mod tests {
     pub struct FakeParentArgs {}
     impl FakeParentArgs {
         pub fn once(args: &str) -> Self {
-            Self::new(args, |v| TlsState::Once(v), "once")
+            Self::new(args, TlsState::Once, "once")
         }
         pub fn for_scope(args: &str) -> Self {
-            Self::new(args, |v| TlsState::Scope(v), "for_scope")
+            Self::new(args, TlsState::Scope, "for_scope")
         }
         fn new<F>(args: &str, initial: F, from_: &str) -> Self
         where
@@ -665,7 +686,7 @@ pub mod tests {
             MockProcInfo {
                 delta_pid: processes.last().map(|p| p.0).unwrap_or(1),
                 info: processes
-                    .into_iter()
+                    .iter()
                     .map(|(pid, start_time, cmd, ppid)| {
                         let cmd_vec = cmd.split(' ').map(str::to_owned).collect();
                         (*pid, FakeProc::new(*pid, *start_time, cmd_vec, *ppid))
@@ -700,21 +721,21 @@ pub mod tests {
     #[test]
     fn test_process_testing() {
         {
-            let _args = FakeParentArgs::once(&"git blame hello");
+            let _args = FakeParentArgs::once("git blame hello");
             assert_eq!(
                 calling_process_cmdline(ProcInfo::new(), guess_git_blame_filename_extension),
                 Some("hello".into())
             );
         }
         {
-            let _args = FakeParentArgs::once(&"git blame world.txt");
+            let _args = FakeParentArgs::once("git blame world.txt");
             assert_eq!(
                 calling_process_cmdline(ProcInfo::new(), guess_git_blame_filename_extension),
                 Some("txt".into())
             );
         }
         {
-            let _args = FakeParentArgs::for_scope(&"git blame hello world.txt");
+            let _args = FakeParentArgs::for_scope("git blame hello world.txt");
             assert_eq!(
                 calling_process_cmdline(ProcInfo::new(), guess_git_blame_filename_extension),
                 Some("txt".into())
@@ -730,7 +751,7 @@ pub mod tests {
     #[test]
     #[should_panic]
     fn test_process_testing_assert() {
-        let _args = FakeParentArgs::once(&"git blame do.not.panic");
+        let _args = FakeParentArgs::once("git blame do.not.panic");
         assert_eq!(
             calling_process_cmdline(ProcInfo::new(), guess_git_blame_filename_extension),
             Some("panic".into())
@@ -742,7 +763,7 @@ pub mod tests {
     #[test]
     #[should_panic]
     fn test_process_testing_assert_never_used() {
-        let _args = FakeParentArgs::once(&"never used");
+        let _args = FakeParentArgs::once("never used");
 
         // causes a panic while panicing, so can't test:
         // let _args = FakeParentArgs::for_scope(&"never used");
@@ -751,7 +772,7 @@ pub mod tests {
 
     #[test]
     fn test_process_testing_scope_can_remain_unused() {
-        let _args = FakeParentArgs::for_scope(&"never used");
+        let _args = FakeParentArgs::for_scope("never used");
     }
 
     #[test]
@@ -924,7 +945,7 @@ pub mod tests {
         ]);
         assert_eq!(
             calling_process_cmdline(parent, describe_calling_process),
-            Some(CallingProcess::GitGrep(empty_command_line.clone()))
+            Some(CallingProcess::GitGrep(empty_command_line))
         );
 
         for grep_command in &[
@@ -1003,7 +1024,7 @@ pub mod tests {
                 assert_eq!(cmd_line.short_options, set(&["-w"]));
                 assert_eq!(ext, Some(expected_extension.to_string()));
             } else {
-                assert!(false);
+                unreachable!();
             }
         }
     }
